<'
# Copyright (c) 2012-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Thu Aug  2 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <cerrno>
#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <vector>

#include "ros/master.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "private.h"


/* <"[--- ${comp}_init --------------------------------------------------]"> */

genom_client
genom_<"$comp">_client_init(int argc, char *argv[])
{
  static char progname[] = "<"$comp">";
  std::vector<std::string> nodes;
  pthread_mutexattr_t attr;
  struct ros_options opts;
  char *largv[argc+1];
  genom_state_component state;
  genom_client h;
  int i, e;

  h = new genom_client_s;
  if (!h) return NULL;
  h->instance = NULL;
  h->node = NULL;
  h->connections = NULL;
  h->events = 0;
  h->next = 0;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init(&h->activelk, &attr);

  /* must shift args so that getopt() works (argv[0] must be progname) */
  largv[0] = progname;
  for(i=0; i<argc; i++) {
    largv[i+1] = argv[i];
  }

  h->error = ros_getopt(argc+1, largv, &opts);
  if (h->error) {
    free(h);
    return NULL;
  }

  h->error = ros_init(argc+1, largv);
  if (h->error) goto error;

  h->instance = strdup(opts.instance?opts.instance:"<"$comp">");
  try {
    /* lookup server */
    if (!ros::master::getNodes(nodes)) {
      genom_mwerr_detail d;
      snprintf(d.what, sizeof(d.what), "cannot list running nodes");
      h->error = genom_mwerr(&d);
      goto error;
    }
    for(i=e=0; (unsigned)i<nodes.size(); i++) {
      if (!strcmp(h->instance, nodes[i].c_str()+1) && nodes[i][0] == '/') {
        e = 1;
        break;
      }
    }
    if (!e) {
      genom_mwerr_detail d;
      snprintf(d.what, sizeof(d.what),
               "%s node does not seem to be running", h->instance);
      h->error = genom_mwerr(&d);
      goto error;
    }

    h->node = new ros::NodeHandle;
  }
  catch (ros::Exception e) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what), "%s", e.what());
    h->error = genom_mwerr(&d);
    goto error;
  }

  /* check digest */
  h->error = genom_<"$comp">_client_genom_state_port(h, &state);
  if (h->error) goto error;

  if (strcmp(state.digest, "<"[$component digest]">")) {
    genom_incompatible_digest_detail d;
    strcpy(d.server.version, state.version);
    strcpy(d.server.date, state.date);
    strcpy(d.client.version, "<"[$component version]">");
    strcpy(d.client.date, "<"[clock format [clock sec]]">");
    genom_incompatible_digest(&d);
    goto error;
  }

  /* connect to services */
  h->connections = new struct connections;
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
  h->connections-><"[$s name]"> = h->node->serviceClient<
    genom::srv::<"[$s name]">::input, genom::srv::<"[$s name]">::output
    >("/" + std::string(h->instance) + "/<"[$s name]">", true);
  if (!h->connections-><"[$s name]">.exists()) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what),
             "cannot connect to service %s", "<"[$s name]">");
    h->error = genom_mwerr(&d);
    goto error;
  }
<'  } else {'>
  h->connections-><"[$s name]"> =
    new gnmix_<"[$s name]">(
      *h->node, "/" + std::string(h->instance) + "/<"[$s name]">");
  if (!h->connections-><"[$s name]">->waitForActionServerToStart(
        ros::Duration(1.))) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what),
             "cannot connect to action %s", "<"[$s name]">");
    h->error = genom_mwerr(&d);
  }
<'  }'>
<'}'>

  return h;

error:
  if (h->connections) {
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
    h->connections-><"[$s name]">.shutdown();
<'  } else {'>
    if (h->connections-><"[$s name]">)
      delete h->connections-><"[$s name]">;
<'  }'>
<'}'>
    delete h->connections;
  }
  if (h->instance) free(h->instance);
  if (h->node) delete h->node;
  pthread_mutex_destroy(&h->activelk);
  delete h;
  ros_fini();
  return NULL;
}


/* <"[--- ${comp}_fini --------------------------------------------------]"> */

void
genom_<"$comp">_client_fini(genom_client h)
{
  if (h) {
    if (h->connections) {
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
      h->connections-><"[$s name]">.shutdown();
<'  } else {'>
      if (h->connections-><"[$s name]">)
        delete h->connections-><"[$s name]">;
<'  }'>
<'}'>
      delete h->connections;
    }
    if (h->instance) free(h->instance);
    if (h->node) delete h->node;
    pthread_mutex_destroy(&h->activelk);
    delete h;
    ros_fini();
  }
}


/* <"[--- ${comp}_name --------------------------------------------------]"> */

const char *
genom_<"$comp">_client_instance(genom_client h)
{
  return h->instance;
}


/* <"[--- ${comp}_eventfd -----------------------------------------------]"> */

int
genom_<"$comp">_client_eventfd(genom_client h)
{
  return ros_eventfd();
}


/* <"[--- ${comp}_service_info ------------------------------------------]"> */

const struct genom_service_info *
genom_<"$comp">_client_service_info(genom_client h, int rqstid)
{
  return ros_service_info(h, rqstid);
}


/* <"[--- ${comp}_done --------------------------------------------------]"> */

int
genom_<"$comp">_client_done(genom_client h, int rqstid)
{
  return ros_done(h, rqstid);
}


/* <"[--- ${comp}_wait --------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_wait(genom_client h, int rqstid)
{
  return ros_waitterm(h, rqstid);
}


/* <"[--- ${comp}_clean -------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_clean(genom_client h, int rqstid)
{
  return ros_clean(h, rqstid);
}


/* <"[--- ${comp}_doevents -----------------------------------------------]"> */

genom_event
genom_<"$comp">_client_doevents(genom_client h)
{
  return ros_doevents(h);
}
